

set -ex



pip check
python -c "import pandas; pandas.test(extra_args=['-m not clipboard and not single_cpu and not slow and not network and not db and not arm_slow', '-n=2', '-k', 'not (_not_a_real_test or test_rolling_var_numerical_issues or test_std_timedelta64_skipna_false or test_value_counts_normalized[M8[ns]] or test_to_datetime_format_YYYYMMDD_with_nat or (TestReductions and test_median_2d))'])"
exit 0
